import os

from django.core.management import CommandError
from django.core.management.templates import TemplateCommand

from jackal.settings import jackal_settings


class Command(TemplateCommand):
    """
    jackal settings 내에 지정한 APP_DIR 내로 앱을 생성합니다.
    """

    help = (
        "Create app in app folder"
    )
    missing_args_message = "You must provide an application name."

    create_file = {
        'serializer.py': 'from rest_framework import serializers\n\n',
    }
    remove_file = [
        'tests.py', 'admin.py', 'views.py',
    ]

    def handle(self, **options):
        app_name = options.pop('name')

        app_root = jackal_settings.APP_DIR

        try:
            if app_root is not None:
                app_path = os.path.join(app_root, app_name)
                try:
                    os.mkdir(app_path)
                except FileExistsError:
                    raise CommandError("'%s' already exists" % app_name)
            else:
                app_path = None

            super().handle('app', app_name, app_path, **options)
        except CommandError as e:
            raise e

        if app_path is None:
            app_path = os.path.join(os.getcwd(), app_name)

        for file in self.remove_file:
            os.remove(app_path + '/' + file)

        for file_name, inner_data in self.create_file.items():
            with open(app_path + '/{}'.format(file_name), 'w') as f:
                f.write(inner_data)

        return



from rest_framework.response import Response
from rest_framework.views import APIView

from jackal.filters import JackalQueryFilter
from jackal.helpers.data_helper import JackalDictMapper
from jackal.inspectors import Inspector
from jackal.paginators import JackalPaginator
from jackal.settings import jackal_settings


class _Response:
    def success(self, detail='success', **kwargs):
        return Response({'detail': detail, **kwargs})

    def simple_response(self, result=None, status=200, headers=None, **kwargs):
        return Response(result, status=status, headers=headers, **kwargs)

    def bad_request(self, **data):
        return Response(data, status=400)

    def forbidden(self, **data):
        return Response(data, status=403)

    def internal_server_error(self, **data):
        return Response(data, status=500)


class _Getter:
    def get_queryset(self):
        return self.queryset

    def get_model(self):
        if self.model is not None:
            return self.model
        else:
            return self.get_queryset().model

    def get_lookup_map(self, **additional):
        d = self.lookup_map
        return {**d, **additional}

    def get_filter_map(self, **additional):
        d = self.filter_map
        return {**d, **additional}

    def get_search_dict(self, **additional):
        d = self.search_dict
        return {**d, **additional}

    def get_extra_kwargs(self, **additional):
        d = self.extra_kwargs
        return {**d, **additional}

    def get_query_filter_class(self):
        if self.query_filter is None:
            return JackalQueryFilter
        return self.query_filter

    def get_serializer_class(self):
        return self.serializer_class

    def get_serializer_context(self, **kwargs):
        return kwargs

    def get_jackal_exception_handler(self):
        return jackal_settings.EXCEPTION_HANDLER

    def get_inspector(self, request):
        inspect_map = self.get_cur_inspect_map()
        if not inspect_map:
            return None
        ins_class = self.inspector_class
        return ins_class(request.data, inspect_map)

    def get_cur_inspect_map(self):
        return getattr(self, '{}_inspect_map'.format(self.request.method), self.inspect_map)

    def get_user_field(self):
        return self.user_field

    def get_permission_classes(self):
        return self.default_permission_classes + self.permission_classes

    def get_authentication_classes(self):
        return self.default_authentication_classes + self.authentication_classes

    def get_authenticators(self):
        return [auth() for auth in self.get_authentication_classes()]

    def get_permissions(self):
        return [permission() for permission in self.get_permission_classes()]


class _PrePost:
    def pre_method_call(self, request, *args, **kwargs):
        """
        This will call before method run. Likes get, post, patch...
        """
        pass

    def post_method_call(self, request, response, *args, **kwargs):
        """
        This will call after method run. Likes get, post, patch...
        """
        pass

    def pre_check_permissions(self, request):
        pass

    def post_check_permissions(self, request):
        pass

    def pre_check_object_permissions(self, request, obj):
        pass

    def post_check_object_permissions(self, request, obj):
        pass

    def pre_handle_exception(self, exc):
        pass


class JackalBaseAPIView(_Getter, _PrePost, _Response, APIView):
    default_permission_classes = ()
    default_authentication_classes = ()
    permission_classes = ()
    authentication_classes = ()

    model = None
    queryset = None

    lookup_map = {}
    filter_map = {}
    search_dict = {}
    extra_kwargs = {}
    inspect_map = {}
    user_field = ''

    default_page_number = 1
    default_page_length = None

    search_keyword_key = 'search_keyword'
    search_type_key = 'search_type'
    page_number_key = 'page_number'
    page_length_key = 'page_length'
    ordering_key = 'ordering'

    serializer_class = None
    query_filter = JackalQueryFilter
    inspector_class = Inspector
    paginator_class = JackalPaginator

    def dispatch(self, request, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        request = self.initialize_request(request, *args, **kwargs)
        self.request = request
        self.headers = self.default_response_headers

        try:
            self.initial(request, *args, **kwargs)
            if request.method.lower() in self.http_method_names:
                handler = getattr(self, request.method.lower(),
                                  self.http_method_not_allowed)
            else:
                handler = self.http_method_not_allowed
            self.pre_method_call(request, *args, **kwargs)
            response = handler(request, *args, **kwargs)
            self.post_method_call(request, response, *args, **kwargs)

        except Exception as exc:
            response = self.handle_exception(exc)

        self.response = self.finalize_response(request, response, *args, **kwargs)

        return self.response

    def handle_exception(self, exc):
        """
        high jacking exception and handle with jackal_exception_handler
        """
        self.pre_handle_exception(exc)
        jackal_handler = self.get_jackal_exception_handler()
        context = self.get_exception_handler_context()
        response = jackal_handler(exc, context)
        if response is not None:
            response.exception = True
            return response
        else:
            return super().handle_exception(exc)

    def check_permissions(self, request):
        self.pre_check_permissions(request)
        super().check_permissions(request)
        self.post_check_permissions(request)

    def check_object_permissions(self, request, obj):
        self.pre_check_object_permissions(request, obj)
        super().check_object_permissions(request, obj)
        self.post_check_object_permissions(request, obj)

    def get_object(self, request, **kwargs):
        queryset = self.get_user_queryset(queryset=self.get_model().objects.all(), request=request)

        f_class = self.get_query_filter_class()
        f = f_class(queryset=queryset, params=request.query_params)

        lookup_map = self.get_lookup_map()
        extra_kwargs = self.get_extra_kwargs(**JackalDictMapper.av2bv(lookup_map, kwargs))

        obj = f.extra(**extra_kwargs).get(raise_404=True)
        self.check_object_permissions(request=request, obj=obj)
        return obj

    def get_filtered_queryset(self, request, **kwargs):
        queryset = self.get_user_queryset(queryset=self.get_queryset(), request=request)

        f_class = self.get_query_filter_class()
        f = f_class(queryset=queryset, params=request.query_params)

        lookup_map = self.get_lookup_map()
        filter_map = self.get_filter_map()
        search_dict = self.get_search_dict()
        extra_kwargs = self.get_extra_kwargs()
        extra_kwargs.update(JackalDictMapper.av2bv(lookup_map, kwargs))

        queryset = f.search(
            search_dict,
            search_keyword_key=self.search_keyword_key,
            search_type_key=self.search_type_key
        ).filter_map(filter_map).extra(
            **extra_kwargs
        ).ordering(
            self.ordering_key
        ).queryset.distinct()
        return queryset

    def get_user_queryset(self, request, queryset=None):
        if queryset is None:
            queryset = self.get_queryset()

        user_field = self.get_user_field()
        if user_field:
            queryset = queryset.filter(**{user_field: request.user})

        return queryset

    def get_inspected_data(self, request):
        inspector = self.get_inspector(request)
        if inspector is not None:
            return inspector.inspected_data
        return dict()

    def get_paginator_class(self):
        return self.paginator_class

    def get_response_data(self, request, **kwargs):
        pass

    def get_paginated_data(self, request, queryset):
        page_num = request.query_params.get(self.page_number_key, self.default_page_number)
        page_length = request.query_params.get(self.page_length_key, self.default_page_length)

        paginator_class = self.get_paginator_class()
        paginator = paginator_class(queryset, page_num, page_length)

        paginated_data = paginator.serialized_data(
            serializer_class=self.get_serializer_class(),
            context=self.get_serializer_context()
        )
        return paginated_data


class JackalAPIView(JackalBaseAPIView):
    pass



from rest_framework.exceptions import APIException
from rest_framework.response import Response


class JackalAPIException(APIException):
    default_message = ''
    status_code = None

    def __str__(self):
        return self.__class__.__name__

    def response_data(self):
        return {}


class MessageException(JackalAPIException):
    default_message = ''

    def __init__(self, message=None, context=None, **kwargs):
        self.message = message if message is not None else self.default_message
        self.context = context if context is not None else dict()
        self.kwargs = kwargs

    def response_data(self):
        return {'message': self.message, **self.kwargs}


class NotFound(MessageException):
    status_code = 404
    message_form = 'Can not find {} model instance filtered by \'{}\''

    def __init__(self, model, filters, context=None, **kwargs):
        self.model = model
        self.filters = filters
        self.context = context
        self.kwargs = kwargs

    @property
    def message(self):
        filter_condition = ', '.join(['{}={}'.format(key, value) for key, value in self.filters.items()])
        return self.message_form.format(self.model.__name__, filter_condition)

    def response_data(self):
        return {'message': self.message, 'model': self.model.__name__, **self.kwargs}


class Forbidden(MessageException):
    status_code = 403


class BadRequest(MessageException):
    status_code = 400


class FieldException(JackalAPIException):
    status_code = 400

    def __init__(self, field, message, context=None, **kwargs):
        self.field = field
        self.message = message
        self.kwargs = kwargs
        self.context = context if context is not None else dict()

    def response_data(self):
        return {
            'message': self.message,
            'field': self.field,
            **self.kwargs
        }


def jackal_exception_handler(exc, context):
    if isinstance(exc, JackalAPIException):
        return Response(exc.response_data(), status=exc.status_code)
    return None



from rest_framework import permissions

REAL_SAFE = ('OPTIONS', 'HEAD')
POST = 'POST'
GET = 'GET'
PATCH = 'PATCH'
PUT = 'PUT'
DELETE = 'DELETE'


class _HttpMethodPermission(permissions.BasePermission):
    allow_method = ''

    def has_permission(self, request, view):
        if request.method in REAL_SAFE:
            return True
        return request.method == self.allow_method

    def has_object_permission(self, request, view, obj):
        return self.has_permission(request, view)


class IsGet(_HttpMethodPermission):
    """
    GET 요청을 체크합니다.
    """
    allow_method = GET


class IsPost(_HttpMethodPermission):
    """
    POST 요청을 체크합니다.
    """
    allow_method = POST


class IsPut(_HttpMethodPermission):
    """
    PATCH 요청을 체크합니다.
    """
    allow_method = PUT


class IsPatch(_HttpMethodPermission):
    """
    PATCH 요청을 체크합니다.
    """
    allow_method = PATCH


class IsDelete(_HttpMethodPermission):
    """
    DELETE 요청을 체크합니다.
    """
    allow_method = DELETE


class HasAuth(permissions.IsAuthenticated):
    pass


class HasAuthOrReadOnly(permissions.IsAuthenticatedOrReadOnly):
    pass



from jackal.exceptions import NotFound
from jackal.helpers.data_helper import isiter
from jackal.loaders import query_function_loader
from jackal.shortcuts import gen_q


class JackalQueryFilter:
    def __init__(self, queryset, params=None):
        self.queryset = queryset
        self.params = params or {}

    @staticmethod
    def _get_query_function(key):
        for n, callback in query_function_loader().items():
            if key.find(n) >= 0:
                return key.replace('__{}'.format(n), ''), callback
        else:
            return key, None

    def filter_map(self, filter_map):
        """
        params = {
            'name': 'Yongjin',
            'age_lowest': '20',
            'skills': 'django,python',
            'status[]': [1, 2, 3],
        }
        f = JackalQueryFilter(User.objects.all(), params)

        queryset = f.filter_map({
            'name': 'name__contains',
            'age_lowest': 'age__gte',
            'skills__to_list': 'skills__in',
            'status[]': 'status__in',
        }).queryset
        """

        queryset = self.queryset
        # initial
        filterable = {}
        filterable_q_objects = list()

        for map_key, filter_keyword in filter_map.items():
            # eg) map_key       : 'skills__to_list'
            #     filter_keyword: 'skills__in'
            map_key, callback = self._get_query_function(map_key)
            # eg) map_key  : skills
            #     call_back: to_list function at DefaultQueryFunction
            if map_key.find('[]') > 0 and hasattr(self.params, 'getlist'):  # django drf query_params getlist support
                param_value = self.params.getlist(map_key)
                if param_value in [None, '', []]:
                    continue  # if empty, skip
            else:
                param_value = self.params.get(map_key)
            # eg) param_value  : 'django,python'

            if param_value in [None, '']:
                continue
            elif callback is not None:
                param_value = callback(param_value)
                # eg) param_value  : ['django', 'python']

            if isiter(filter_keyword):  # if filter_keyword is such like ('name__contains', 'job__contains').
                # make Q object like Q(Q(name__contains={param_value}) | Q(job__contains={param_value}))
                filterable_q_objects.append(gen_q(param_value, *filter_keyword))
            else:
                filterable[filter_keyword] = param_value

        self.queryset = queryset.filter(**filterable).filter(*filterable_q_objects)
        return self

    def search(self, search_dict, search_keyword_key='search_keyword', search_type_key='search_type'):
        """
        params = {
            'search_keyword': 'Yongjin',
            'search_type': 'name',
        }
        f = JackalQueryFilter(User.objects.all(), params)

        queryset = f.search({
            'all': ('name__contains', 'job__contains', 'city__contains),
            'name': 'name__contains',
            'job': 'job__contains',
            'city': 'city__contains',
        }).queryset
        """

        queryset = self.queryset

        search_keyword = self.params.get(search_keyword_key)
        # if search_type is not exists, consider all type.
        search_type = self.params.get(search_type_key, 'all')
        # get search_type
        dict_value = search_dict.get(search_type, None)

        if dict_value is not None and search_keyword is not None:
            if isiter(dict_value):
                self.queryset = queryset.filter(gen_q(search_keyword, *dict_value))
            else:
                self.queryset = queryset.filter(**{dict_value: search_keyword})
        return self

    def extra(self, **extra_kwargs):
        """
        f = JackalQueryFilter(User.objects.all(), {})
        queryset = f.extra(age__lte=30, is_active=True).queryset
        """

        queryset = self.queryset
        self.queryset = queryset.filter(**extra_kwargs)
        return self

    def ordering(self, ordering_key='ordering'):
        """
        params = {
            'ordering': 'name,-age',
        }
        f = JackalQueryFilter(User.objects.all(), params)

        queryset = f.ordering().queryset
        """

        ordering = self.params.get(ordering_key)
        if ordering:
            queryset = self.queryset
            order_by = ordering.split(',')
            self.queryset = queryset.order_by(*order_by)
        return self

    def get(self, raise_404=False, **kwargs):
        """
        f = JackalQueryFilter(User.objects.all(), {})
        user = f.get(id=5)
        """
        queryset = self.queryset
        obj = self.queryset.filter(**kwargs).first()
        if obj is None and raise_404:
            raise NotFound(model=queryset.model, filters=kwargs)

        return obj
